#!/usr/bin/env python2
from __future__ import print_function

"""filedrop.py: http upload target"""
__version__   = "0.9"
__author__    = "ed <filedrop@ocv.me>"
__url__       = "https://ocv.me/dev/?filedrop.py"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2019

# for a smaller (and much slower) alternative,
#   see https://ocv.me/dev/?filesink.py
#
# expect 210 MB/s with python3, or 285 MB/s using python2,
# that's like 9x faster than python's http parser at 32 MB/s
# (cgi.FieldStorage, which is what filesink.py uses)
#
# navigate to your ip:port combination with a browser to see a
# basic upload form, or just use curl to upload windows7.iso:
#   curl -F "f=@windows7.iso" http://127.0.0.1:1234/
#
# the uploader will receieve a minimal html response with the
# filesize and sha1 checksum of the uploaded file, as seen by us


import os
import sys
import time
import hashlib
import select

try:
	from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
except:
	from http.server import BaseHTTPRequestHandler, HTTPServer


class Srv(BaseHTTPRequestHandler):
	def readbytes(self, f, num_bytes=4096):
		"""blocks until any data, returns [] on disconnect"""
		while True:
			r,w,e = select.select([f], [], [], 0)
			if f in r:
				return os.read(f.fileno(), num_bytes)
			
			time.sleep(0.01)
	
	def readline(self, f):
		"""reads one ascii \r\n line aggressively"""
		ret = b''
		while True:
			ret += self.readbytes(f, 2)
			if ret[-1:] == b'\r':
				ret += self.readbytes(f, 1)
			if ret[-2:] == b'\r\n':
				return ret[:-2]
	
	def reply(self, body):
		header = 'HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: {0}\r\n\r\n'.format(len(body)).encode('ascii')
		self.wfile.write(header + body)
		return body
	
	def loud_reply(self, body):
		print(body.decode('ascii'), end='')
		self.reply(body)
	
	def do_GET(self):
		print("\nGET  {0}".format(self.client_address))
		self.reply(b'<form method="post" enctype="multipart/form-data"><input type="file" name="f"><input type="submit">')
	
	def extract_filedata(self, boundary, ifd):
		blen = len(boundary)
		while True:
			buf = self.readbytes(ifd)
			if not buf:
				# abort: client disconnected
				print("CLIENT DISCONNECT 1")
				return
			
			while True:
				d = len(buf) - blen
				if d > 0:
					# buffer growing large; yield everything except
					# the part at the end (maybe start of boundary)
					yield buf[:d]
					buf = buf[d:]
				
				# look for boundary near the end of the buffer
				for n in range(1, len(buf)+1):
					if not buf[-n:] in boundary:
						n -= 1
						break
				
				if n == 0 \
				or not boundary.startswith(buf[-n:]):
					# no boundary contents near the buffer edge
					break
				
				if blen == n:
					# EOF: found boundary
					yield buf[:-n]
					return
				
				buf2 = self.readbytes(ifd)
				if not buf2:
					# abort: client disconnected
					print("CLIENT DISCONNECT 2")
					return
				
				buf += buf2
				
			yield buf
	
	def handle_multipart(self, boundary, ifd, ofd):
		tlen = 0
		hashobj = hashlib.sha1()
		for buf in self.extract_filedata(boundary, ifd):
			tlen += len(buf)
			hashobj.update(buf)
			ofd.write(buf)
		
		return tlen, hashobj.hexdigest()
	
	def do_POST(self):
		print("\nPOST {0}".format(self.client_address))
		#for k, v in self.headers.items():
		#	print('[{0}] [{1}]'.format(k, v))

		try:
			if self.headers['expect'].lower() == '100-continue':
				self.wfile.write(b'HTTP/1.1 100 Continue\r\n\r\n')
		except:
			pass
		
		ln = self.readline(self.rfile)
		boundary = b'\r\n' + ln + b'--\r\n'
		while ln:
			print(ln)
			ln = self.readline(self.rfile)
		
		fn = 'inc.{0:.3f}'.format(time.time())
		print('writing to {0}'.format(fn))
		with open(fn, 'wb') as f:
			sz, sha1 = self.handle_multipart(boundary, self.rfile, f)
		
		msg = '{0} bytes // sha1 {1}\n'.format(sz, sha1)
		self.loud_reply(msg.encode('ascii'))


port = 1234
if len(sys.argv) > 1:
	port = int(sys.argv[1])

print('listening @ http://127.0.0.1:{0}/'.format(port))
HTTPServer(('0.0.0.0', port), Srv).serve_forever()
